      subroutine get_triangulation(infile,tripath,trifile,triflag,
     . debug,list,lptr,lend,x,y,z,azimuth, zenith, milliaz, millizen,n)
C changes 2008-March-06: added tripath to calls to read_triangulation_table
C  and write_triangulation_table.     
C Added zenazstring corresponding to each grid position in INFILE.
C This subroutine either generates a triangulation using Renka's Algorithm 772
C for the grid points contained within INFILE or reads it from tripath/trifile.
C 
C Inputs:
C INFILE = an ascii file with the first line (format I5) containing the number
C          of grid points and the remaining lines (format 2F16.9) containing 
C          the spacecraft elevation and azimuth in degrees for each grid point.
C 	   Note, this file is only read in if triflag > 0.
C TRIPATH = the path to TRIFILE.
C TRIFILE = a fits file containing the triangulation indices LIST,LPTR,LEND,
C           the cartesian coordinates of the grid points, X,Y,Z, and the
C           S/C Azimuth/Zenith of each grid point. This file is read in if
C           TRIFLAG = 0 and written if TRIFLAG = 1.
C TRIFLAG = flag determining whether the triangulation is read in,generated,
C           and/or stored.
C TRIFLAG = 0 - triangulation is read in from TRIPATH/TRIFILE
C TRIFLAG = 1 - triangulation is generated for gridpoints in INFILE and output
C               to TRIPATH/TRIFILE for future use.
C TRIFLAG = 2 - triangulation is generated for gridpoints in INFILE and is not
C               written to a file.
C DEBUG >1 increases verbosity
C Outputs:
C LIST,LPTR, LEND - triangulation indices needed by Renka's TRFIND routine.
C X,Y,Z - cartesian coordinates (S/C) for each grid point. Needed for Renka's
C          TRFIND routine.
C Azimuth, Zenith - S/C coordinates in degrees for each grid point. Needed to
C                   find the corresponding response file.
C ZenAzString - unique string that denotes zenith and azimuth for each drm file
C N - the number of points in the triangulation
      implicit none
      character*100 infile
      character*80 trifile, tripath
      integer triflag, n, list(*), lend(*), lptr(*)
      real x(*), y(*), z(*), azimuth(*), zenith(*)
      integer milliaz(*), millizen(*)
      integer LIN,nmax,debug
      parameter (nmax=1000)
      integer IWK(nmax,2),LNEW,i,k,ier,maz(nmax),mzen(nmax)
      real sf,rlat(nmax),rlon(nmax),wt(nmax)
      SF = ATAN(1.)/45.
C
C Logical unit numbers for input file
      data LIN/1/
C
C Input Formats:
C
  301 FORMAT (I5)
  320 FORMAT (2F16.9,2x,i6,2x,i6)
      IF (TRIFLAG.GT.0) THEN
        OPEN (LIN,FILE=infile,STATUS='OLD')
C
C *** Read the input data 
C
        READ (LIN,301,ERR=160) N
        DO 10 K = 1,N
          READ (LIN,320,ERR=160,END=11) RLAT(K), RLON(K),
     .     mzen(k),maz(k)	  
c	  print*,rlat(k),rlon(k),'  ',zenaz(k)
   10   CONTINUE
   11   WRITE (*,*) 'End of input file ', N,' points read.'
        CLOSE(LIN)
        DO 12 K=N+1,NMAX
          RLAT(K)=0.
	  RLON(K)=0.
	  maz(k)=0
	  mzen(k)=0
   12   CONTINUE	  
  
C *** Convert the nodes (RLAT,RLON) to radians and then
C     to Cartesian coordinates (X,Y,Z) and create the
C     triangulation.
C
        DO 30 K = 1,N
	  ZENITH(K) = 90.-RLAT(K) !needed to find indiv responses
	  AZIMUTH(K) = RLON(K)
          RLAT(K) = SF*RLAT(K)
          RLON(K) = SF*RLON(K)
	  milliaz(k) = maz(k)
	  millizen(k) = mzen(k)
	  if (debug.gt.1) 
     .     print*,k,rlat(k),rlon(k),azimuth(k),zenith(k),
     .	   milliaz(k),millizen(k)
   30   CONTINUE
        
        CALL TRANS (N,RLAT,RLON, X,Y,Z)
        CALL TRMESH (N,X,Y,Z, LIST,LPTR,LEND,LNEW,IWK,
     .             IWK(1,2),WT,IER)
        IF (IER .NE. 0) THEN
          WRITE (*,420) IER
          STOP
        ENDIF
C *** Output the triangulation info if triflag = 1
        IF (triflag.EQ.1) THEN
	  call write_triangulation_hdr0 (trifile, tripath)
	  call write_triangulation_table (tripath,trifile,list,lptr,lend,
     .            x,y,z,azimuth,zenith,milliaz,millizen,n)
        ENDIF !triflag =1 write triangulation to file.
      ELSE ! triflag = 0 read triangulation from file
        call read_triangulation_table(tripath,trifile, list, lptr, lend,
     .    x,y,z,azimuth,zenith,milliaz,millizen,n)
        if(debug.gt.1) then
          write(*,*) 'n = ',n
          do i=1,10
	    write(*,*) lend(i),x(i),y(i),z(i),azimuth(i),zenith(i)
	  enddo
	  do i=n-10,n
	    write(*,*) lend(i),x(i),y(i),z(i),azimuth(i),zenith(i)
	  enddo
	 endif   	
      ENDIF !triflag > 0, so triangulation is to be calculated.
      RETURN
C
C Error message Formats:
C
  160 WRITE (*,410)
      STOP
  410 FORMAT (//5X,'Error encountered in reading data.'/)
  420 FORMAT (//5X,'Error in TRMESH:  IER =',I5/)

      END
